/*
 * matching.c
 *
 * author: Roberto Piassi Passos Bodo
 *  email: rppbodo@gmail.com
 */

#include "matching.h"

/*!
 * This function sends a 'Note Off' event for all the notes that may
 * be sounding.
 */
void all_notes_off() {
	int i, j;
	unsigned char bytes[3];
	Link temp = NULL;

	for (i = 0; i < n_tracks; i++) {
		temp = table[i];

		if (temp != NULL) {
			bytes[0] = temp->bytes[0];
			bytes[2] = 0;
			for (j = 0; j < 128; j++) {
				bytes[1] = j;
				write(output_device_fd, bytes, 3);
			}
		}
	}
}

/*!
 * This function checks if there is any note between the delta time of the
 * current position of each track and new delta time. It returns 1 if there is
 * and 0 if there is not.
 */
int has_accompaniment(Link* links, int delta_time) {
	int i;
	Link temp = NULL;

	for (i = 0; i < n_tracks; i++) {
		temp = links[i];
		if (temp != NULL && temp->delta_time < delta_time) {
			return 1;
		}
	}
	return 0;
}

/*!
 * This function updates all the global variables that store the clocks values.
 */
void update_clocks() {
	long int temp = prev_virtual_clocks[1];
	prev_virtual_clocks[1] = virtual_clock;
	prev_virtual_clocks[0] = temp;

	temp = prev_real_clocks[1];
	prev_real_clocks[1] = real_clock;
	prev_real_clocks[0] = temp;
}

/*!
 * This function gets the current computer time and returns its value in
 * milliseconds.
 */
long int get_current_time() {
	long int milliseconds;
	gettimeofday(&current_time, NULL);
	milliseconds = (current_time.tv_sec - starting_time.tv_sec) * 1000;
	milliseconds += (current_time.tv_usec - starting_time.tv_usec) / 1000;
	return milliseconds;
}

/*!
 * This function handles the execution of the program after a match has
 * been detected.
 */
void after_match(Link link, Link* links) {
	if (!has_accompaniment(links, link->delta_time)) {
		virtual_clock = link->delta_time;
		real_clock = get_current_time();
		update_clocks();
		link = link->next;
	} else {
		FLAG = 1;
		before = virtual_clock;
		after = link->delta_time;
	}
}

/*!
 * This function is the kernel of this program. We can divide it into four parts:
 * (1) first, it handles the musician input in real time;
 * (2) compares it with the score using the LCS problem algorithm;
 * (3) updates the virtual clock according to the musician's tempo;
 * (4) plays the accompaniment given the evolution of the virtual clock.
 */
void matching() {
	struct pollfd fds[2];
	int r, note, errors;
	unsigned char bytes[3];
	Link link = table[soloist_track - 1];
	Link* links = (Link*) malloc(n_tracks * sizeof(Link));

	int i;

	fds[0].fd = input_device_fd;
	fds[0].events = POLLIN;

	fds[1].fd = STDIN_FILENO;
	fds[1].events = POLLIN;

	for (i = 0; i < n_tracks; i++) {
		links[i] = table[i];
	}

	c = (unsigned char**) malloc(2 * sizeof(unsigned char*));

	c[0]
			= (unsigned char*) malloc((2 * MAX_ERRORS + 2)
					* sizeof(unsigned char));

	c[1]
			= (unsigned char*) malloc((2 * MAX_ERRORS + 2)
					* sizeof(unsigned char));

	errors = 0;
	for (i = 0; i < (2 * MAX_ERRORS + 2); i++) {
		c[0][i] = 0;
	}
	c[1][0] = 0;

	virtual_clock = 0;
	real_clock = 0;

	prev_virtual_clocks[0] = -1;
	prev_virtual_clocks[1] = 0;

	prev_real_clocks[0] = -1;
	prev_real_clocks[1] = 0;

	FLAG = 0;

	printf("..::the music has started::..\n");
	printf("(press ESC if you wanna quit)\n");

	gettimeofday(&starting_time, NULL);
	system("/bin/stty cbreak");

	/*
	 * This is the main 'while'.
	 */
	while (link != NULL && link->next != NULL) {
		/*
		 * This FLAG will change the behavior of the execution.
		 */
		if (FLAG == 0) {
			/*
			 * This function looks if there is some input to be read.
			 */
			r = poll(fds, 2, 1);

			/*
			 * (1) If there isn't we have to infer what is the next value of the
			 * virtual clock and this is made by using similarity of triangles.
			 */
			if (r == 0) {
				double opposite_cathetus = prev_virtual_clocks[1]
						- prev_virtual_clocks[0];
				double adjacent_cathetus = prev_real_clocks[1]
						- prev_real_clocks[0];
				long int new_virtual_time;
				real_clock = get_current_time();
				new_virtual_time = opposite_cathetus / adjacent_cathetus;
				new_virtual_time *= (real_clock - prev_real_clocks[0]);
				new_virtual_time += prev_virtual_clocks[0];

				if (new_virtual_time < link->delta_time) {
					virtual_clock = new_virtual_time;
				}
			}
			/*
			 * (2) If there is we have to handle with it.
			 */
			else if (r > 0) {
				if (fds[0].revents & POLLIN) {
					/*
					 * First, we read the input.
					 */
					read(input_device_fd, bytes, 3);

					bytes[0] -= (bytes[0] & 0x0f);
					bytes[0] += (table[soloist_track - 1]->bytes[0] & 0x0f);

					/*
					 * We write the input directly to the output.
					 */
					write(output_device_fd, bytes, 3);

					/*
					 * Then we check if it is a 'Note On' event.
					 */
					if (((bytes[0] & 0xf0) >> 4) == 0x9 && bytes[2] != 0) {
						note = bytes[1];
					} else {
						continue;
					}

					/*
					 * If it is a 'Note On' we can check if there is a match with
					 * the next note in the score.
					 */
					if (note == link->bytes[1] && errors == 0) {
						/*
						 * Here we have a direct match!
						 */
						after_match(link, links);

					}
					/*
					 * Here we do not have a direct match!
					 */
					else {
						Link temp = NULL;

						if (verbose) {
							printf("\nwrong! expecting %d...\n", link->bytes[1]);
						}

						errors++;

						if (errors == MAX_ERRORS) {
							fprintf(stderr,
									"you made 25 mistakes... that's enough!\n");
							all_notes_off();
							break;
						}

						/*
						 * This next 'for' uses the idea of the Longest Common Subsequence
						 * problem to find out if the note is good or not, i.e., if there
						 * is a match.
						 */
						temp = link;
						for (i = 1; i <= (2 * errors + 1) && temp != NULL; i++) {
							if (temp->bytes[1] == note) {
								c[1][i] = c[0][i - 1] + 1;
							} else {
								if (c[0][i] >= c[1][i - 1]) {
									c[1][i] = c[0][i];
								} else {
									c[1][i] = c[1][i - 1];
								}
							}
							temp = temp->next;
						}

						for (i = 1; i <= (2 * errors + 1); i++) {
							if (c[1][i] == (errors / 5) + 1) {
								break;
							}
						}

						/*
						 * Here we have a postponed match!
						 */
						if (i < (2 * errors + 1)) {
							while (i > 1) {
								link = link->next;
								i--;
							}

							after_match(link, links);

							errors = 0;
							for (i = 1; i <= (2 * MAX_ERRORS + 1); i++) {
								c[0][i] = 0;
							}
							c[1][0] = 0;
						}
						/*
						 * Here we do not have a postponed match!
						 */
						else {
							for (i = 1; i <= (2 * errors + 1); i++) {
								c[0][i] = c[1][i];
							}
						}
					}
				} else if (fds[1].revents & POLLIN) {
					/*
					 * We can read also the standard input.
					 */
					char c;
					read(STDIN_FILENO, &c, 1);
					/*
					 * If the user hit the ESC key the program will terminate.
					 */
					if (c == 27) {
						all_notes_off();
						break;
					}
				}

			}
		}
		/*
		 * If the FLAG is set we have to handle with the pending accompaniment.
		 */
		else {
			virtual_clock += ((after - before) / DEADLINE) + 1;

			if (virtual_clock >= after) {
				real_clock = get_current_time();
				update_clocks();
				link = link->next;
				FLAG = 0;
			}
		}

		/*
		 * This is the part that really plays the accompaniment, i.e., that
		 * writes to the output all the notes with delta time smaller than
		 * the virtual clock.
		 */
		for (i = 0; i < n_tracks; i++) {
			if (i != (soloist_track - 1)) {
				while (links[i] != NULL && (links[i]->delta_time
						<= virtual_clock)) {
					write(output_device_fd, links[i]->bytes, 3);
					links[i] = links[i]->next;
				}
			}
		}
	}
	system("/bin/stty -cbreak");

	finish_table();
}
