/*
 * table.c
 *
 * author: Roberto Piassi Passos Bodo
 *  email: rppbodo@gmail.com
 */

#include "table.h"

/*!
 * This function allocates memory for three structures:
 * (1) the list of tempo changes;
 * (2) the main table of the program;
 * (3) a temporary structure called 'events'.
 */
void init_table() {
	int i;

	head_t = NULL;

	table = (Link*) malloc(n_tracks * sizeof(Link));
	for (i = 0; i < n_tracks; i++) {
		table[i] = NULL;
	}

	events = (int*) malloc(n_tracks * sizeof(int));
	for (i = 0; i < n_tracks; i++) {
		events[i] = 0;
	}
}

/*!
 * This function inserts an element into the list of tempo changes.
 */
void insert_tempo(int delta_time, int tempo) {
	Link_t temp, new = (Link_t) malloc(sizeof(Node_t));

	new->delta_time = delta_time;
	new->tempo = tempo;
	new->next = NULL;

	if (head_t == NULL) {
		head_t = new;
	}
	else {
		temp = head_t;

		while (temp->next != NULL) {
			temp = temp->next;
		}

		temp->next = new;
	}
}

/*!
 * This function inserts an element into the table that stores the score.
 */
void insert_table(int track, int delta_time, unsigned char* bytes) {
	Link temp, new = (Link) malloc(sizeof(Node));

	new->delta_time = delta_time;
	memcpy(new->bytes, bytes, 3);
	new->next = NULL;

	if (table[track - 1] == NULL) {
		table[track - 1] = new;
	}
	else {
		temp = table[track - 1];

		while (temp->next != NULL) {
				temp = temp->next;
		}

		temp->next = new;
	}

	events[track - 1] += 1;
}

/*!
 * This function prints the list of tempo changes.
 */
void print_tempo() {
	Link_t temp = head_t;

	while (temp != NULL) {
		printf("(%ld) %d\n", temp->delta_time, temp->tempo);
		temp = temp->next;
	}
	printf("\n");
}

/*!
 * This function prints just one line of the table that stores the score.
 */
void print_table(int track) {
	int i;
	Link temp = NULL;

	for (temp = table[track - 1]; temp != NULL; temp = temp->next) {
		printf("(%ld)", temp->delta_time);
		for (i = 0; i < 3; i++) {
			printf(" %d", temp->bytes[i]);
		}
		printf("\n");
	}
}

/*!
 * This function cleans the track that store the soloist track. It removes
 * all the events different than 'Note On'.
 */
int fix_soloist_track() {
	int note_on, note_off, other;
	int before, after;
	Link temp = NULL;

	/* before cleaning */
	note_on = note_off = other = 0;
	for (temp = table[soloist_track - 1]; temp != NULL; temp = temp->next) {
		if ((temp->bytes[0] >> 4) == 9) {
			if (temp->bytes[2] != 0) {
				note_on += 1;
			} else {
				note_off += 1;
			}
		} else if ((temp->bytes[0] >> 4) == 8) {
			note_off += 1;
		} else {
			other += 1;
		}
	}

	if (note_on != note_off) {
		fprintf(stderr, "error: inconsistent number of notes...\n");
		return 1;
	}

	before = note_on;

	/* cleaning */
	while (table[soloist_track - 1] != NULL) {
		if ((table[soloist_track - 1]->bytes[0] >> 4) == 9) {
			if (table[soloist_track - 1]->bytes[2] == 0) {
				table[soloist_track - 1] = table[soloist_track - 1]->next;
			} else {
				break;
			}
		} else {
			table[soloist_track - 1] = table[soloist_track - 1]->next;
		}
	}

	temp = table[soloist_track - 1];
	while (temp != NULL && temp->next != NULL) {
		if ((temp->next->bytes[0] >> 4) == 9) {
			if (temp->next->bytes[2] == 0) {
				Link aux = temp->next;
				temp->next = aux->next;
				free(aux);
			} else {
				temp = temp->next;
			}
		} else {
			Link aux = temp->next;
			temp->next = aux->next;
			free(aux);
		}
	}

	/* after cleaning */
	note_on = note_off = other = 0;
	for (temp = table[soloist_track - 1]; temp != NULL; temp = temp->next) {
		if ((temp->bytes[0] >> 4) == 9) {
			if (temp->bytes[2] != 0) {
				note_on += 1;
			} else {
				note_off += 1;
			}
		} else if ((temp->bytes[0] >> 4) == 8) {
			note_off += 1;
		} else {
			other += 1;
		}
	}

	if (note_off != 0 || other != 0) {
		fprintf(stderr, "error: inconsistent number of events...\n");
		return 1;
	}

	after = note_on;

	if (before != after) {
		fprintf(stderr, "error: inconsistent number of notes...\n");
		return 1;
	}

	return 0;
}

/*!
 * This function deallocates the memory of three structures:
 * (1) the list of tempo changes;
 * (2) the main table of the program;
 * (3) a temporary structure called 'events'.
 */
void finish_table() {
	int i;

	Link_t temp_t = head_t;
	Link_t aux_t;

	Link temp;
	Link aux;

	while (temp_t != NULL) {
		aux_t = temp_t;
		temp_t = temp_t->next;
		free(aux_t);
	}
	free(head_t);

	for (i = 0; i < n_tracks; i++) {
		temp = table[i];

		while (temp != NULL) {
			aux = temp;
			temp = temp->next;
			free(aux);
		}
		free(table[i]);
	}

	free(events);
}
